<div id="shop-id-label_about">
<h3 class="article-head-title">{{page.title}}</h3>

{% paginate blog.articles by 20 %}

  {% for article in blog.articles  %}
    <div class="article">
      <h3 class="article-head-title">
        <a href="{{article.url}}">{{ article.title }}</a>
      </h3>

      <p>
      {% if blog.comments_enabled? %}
        <a href="{{article.url}}#comments">{{ article.comments_count }} comments</a>
        &mdash;
      {% endif %}
      posted {{ article.created_at | date: "%Y %h" }} by {{ article.author }}</p>
      <div class="article-body textile">
        {{ article.content }}
      </div>
    </div>
  {% endfor %}

  <div id="pagination">
    {{ paginate | default_pagination }}
  </div>

{% endpaginate %}


</div>
<div class="clear-me"></div>
